/*
 * Common clock functionality.
 *
 * Copyright (C) 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/hw_mmio_register_block.h>

#include "cpm-common.h"
#include <math.h>



// Register access.

Cpm_regs::Cpm_regs(l4_addr_t addr, Clock_base *clocks[],
                   uint32_t exclk_freq)
: _clocks(clocks), exclk_freq(exclk_freq)
{
  _regs = new Hw::Mmio_register_block<32>(addr);
}

// Utility methods.

uint32_t
Cpm_regs::get_field(uint32_t reg, uint32_t mask, uint8_t shift)
{
  return (_regs[reg] & (mask << shift)) >> shift;
}

void
Cpm_regs::set_field(uint32_t reg, uint32_t mask, uint8_t shift, uint32_t value)
{
  _regs[reg] = (_regs[reg] & (~(mask << shift))) | ((mask & value) << shift);
}

Clock_base *
Cpm_regs::get_clock(int num)
{
  return _clocks[num];
}



// Field methods.

uint32_t
Field::get_field(Cpm_regs &regs)
{
  if (defined)
    return regs.get_field(reg, mask, bit);
  else
    return 0;
}

void
Field::set_field(Cpm_regs &regs, uint32_t value)
{
  if (defined)
    regs.set_field(reg, mask, bit, value);
}

// Undefined field.

Field Field::undefined;



// Clock sources.

enum Clock_identifiers
Mux::get_input(int num)
{
  if (num < _num_inputs)
    return _inputs[num];
  else
    return Clock_undefined;
}

// Clock sources.

uint8_t
Source::get_source(Cpm_regs &regs)
{
  if (_source.is_defined())
    return _source.get_field(regs);
  else
    return 0;
}

void
Source::set_source(Cpm_regs &regs, uint8_t source)
{
  if (!_source.is_defined())
    return;

  _source.set_field(regs, source);
}

// Clock source frequencies.

uint32_t
Source::get_frequency(Cpm_regs &regs)
{
  // Return the external clock frequency without any input clock.

  if (get_number() == 0)
    return regs.exclk_freq;

  // Clocks with one source yield that input frequency.

  else if (get_number() == 1)
    return regs.get_clock(get_input(0))->get_frequency(regs);

  // With multiple sources, obtain the selected source for the clock.

  uint8_t source = get_source(regs);
  enum Clock_identifiers input = get_input(source);

  // Return the frequency of the source.

  if (input != Clock_undefined)
    return regs.get_clock(input)->get_frequency(regs);
  else
    return 0;
}

// Undefined source.

Source Source::undefined;



// Clock control.

int
Clock_base::have_clock(Cpm_regs &regs)
{
  (void) regs;
  return true;
}

void
Clock_base::start_clock(Cpm_regs &regs)
{
  (void) regs;
}

void
Clock_base::stop_clock(Cpm_regs &regs)
{
  (void) regs;
}

// Default divider.

uint32_t
Clock_base::get_divider(Cpm_regs &regs)
{
  (void) regs;
  return 1;
}

void
Clock_base::set_divider(Cpm_regs &regs, uint32_t division)
{
  (void) regs;
  (void) division;
}

// Clock sources.

uint8_t
Clock_base::get_source(Cpm_regs &regs)
{
  return _source.get_source(regs);
}

void
Clock_base::set_source(Cpm_regs &regs, uint8_t source)
{
  _source.set_source(regs, source);
}

// Clock source frequencies.

uint32_t
Clock_base::get_source_frequency(Cpm_regs &regs)
{
  return _source.get_frequency(regs);
}

// Output clock frequencies.

uint32_t
Clock_base::get_frequency(Cpm_regs &regs)
{
  return get_source_frequency(regs) / get_divider(regs);
}



// PLL-specific control.

int
Pll::have_pll(Cpm_regs &regs)
{
  return _stable.get_field(regs);
}

int
Pll::pll_enabled(Cpm_regs &regs)
{
  return _enable.get_field(regs);
}

int
Pll::pll_bypassed(Cpm_regs &regs)
{
  return _bypass.get_field(regs);
}

// Clock control.

int
Pll::have_clock(Cpm_regs &regs)
{
  return have_pll(regs) && pll_enabled(regs);
}

void
Pll::start_clock(Cpm_regs &regs)
{
  _enable.set_field(regs, 1);
  while (!have_pll(regs));
}

void
Pll::stop_clock(Cpm_regs &regs)
{
  _enable.set_field(regs, 0);
  while (have_pll(regs));
}

// Feedback (13-bit) multiplier.

uint16_t
Pll::get_multiplier(Cpm_regs &regs)
{
  return _multiplier.get_field(regs) + 1;
}

void
Pll::set_multiplier(Cpm_regs &regs, uint16_t multiplier)
{
  _multiplier.set_field(regs, multiplier - 1);
}

// Input (6-bit) divider.

uint8_t
Pll::get_input_division(Cpm_regs &regs)
{
  return _input_division.get_field(regs) + 1;
}

void
Pll::set_input_division(Cpm_regs &regs, uint8_t divider)
{
  _input_division.set_field(regs, divider - 1);
}

// Output (dual 3-bit) dividers.

uint8_t
Pll::get_output_division(Cpm_regs &regs)
{
  uint8_t d0 = _output_division0.get_field(regs);
  uint8_t d1 = _output_division1.get_field(regs);

  return d0 * d1;
}

void
Pll::set_output_division(Cpm_regs &regs, uint8_t divider)
{
  // Assert 1 as a minimum.
  // Divider 0 must be less than or equal to divider 1.

  uint8_t d0 = (uint8_t) floor(sqrt(divider ? divider : 1));
  uint8_t d1 = divider / d0;

  _output_division0.set_field(regs, d0);
  _output_division1.set_field(regs, d1);
}

uint32_t
Pll::get_frequency(Cpm_regs &regs)
{
  // Test for PLL enable and not PLL bypass.

  if (pll_enabled(regs))
  {
    if (!pll_bypassed(regs))
      return (get_source_frequency(regs) * get_multiplier(regs)) /
             (get_input_division(regs) * get_output_division(regs));
    else
      return get_source_frequency(regs);
  }
  else
    return 0;
}

void
Pll::set_pll_parameters(Cpm_regs &regs, uint16_t multiplier, uint8_t in_divider, uint8_t out_divider)
{
  set_multiplier(regs, multiplier);
  set_input_division(regs, in_divider);
  set_output_division(regs, out_divider);

  if (pll_enabled(regs) && !pll_bypassed(regs))
    while (!have_pll(regs));
}



// Clock control.

void
Clock::change_disable(Cpm_regs &regs)
{
  if (_change_enable.is_defined())
    _change_enable.set_field(regs, 0);
}

void
Clock::change_enable(Cpm_regs &regs)
{
  if (_change_enable.is_defined())
    _change_enable.set_field(regs, 1);
}

int
Clock::have_clock(Cpm_regs &regs)
{
  if (_gate.is_defined())
    return !_gate.get_field(regs);
  else
    return true;
}

void
Clock::start_clock(Cpm_regs &regs)
{
  if (_gate.is_defined())
    _gate.set_field(regs, 0);
}

void
Clock::stop_clock(Cpm_regs &regs)
{
  if (_gate.is_defined())
    _gate.set_field(regs, 1);
}

void
Clock::wait_busy(Cpm_regs &regs)
{
  if (_busy.is_defined())
    while (_busy.get_field(regs));
}



// Clock dividers.

uint32_t
Clock::get_divider(Cpm_regs &regs)
{
  if (_divider.is_defined())
    return _divider.get_field(regs) + 1;
  else
    return 1;
}

void
Clock::set_divider(Cpm_regs &regs, uint32_t division)
{
  if (!_divider.is_defined())
    return;

  change_enable(regs);
  _divider.set_field(regs, division - 1);
  wait_busy(regs);
  change_disable(regs);
}

void
Clock::set_source(Cpm_regs &regs, uint8_t source)
{
  change_enable(regs);
  Clock_base::set_source(regs, source);
  wait_busy(regs);
  change_disable(regs);
}
